//---------------------------------------------------------------------------
#ifndef __BORLANDC__
	// VC++ needs to know that it must link setupapi.lib
	#pragma comment(lib, "setupapi.lib")
	//  Use the -tWM  (32-bit multi-threaded target) command-line switch
	#pragma checkoption -tWM
#endif

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <setupapi.h>
#include <process.h>

#include "usb.h"
#include "commands.h"
#include "errmsgs.h"

#define ILCD_VENDOR_ID		0x0483
#define ILCD_PRODUCT_ID		0x1311
#define USB_MAX_OUT_REPORT_SIZE	65		// output report size will never exceed 65 bytes
#define USB_MAX_IN_REPORT_SIZE	65    // input report size will never exceed 65 bytes
#define HIDP_STATUS_SUCCESS  0
#define USB_NUM_BUFFERS		32

typedef struct _HIDD_ATTRIBUTES
{
	 ulong Size; // = sizeof (struct _HIDD_ATTRIBUTES)

	 ushort VendorID;
	 ushort ProductID;
	 ushort VersionNumber;

} HIDD_ATTRIBUTES, *PHIDD_ATTRIBUTES;

typedef struct _HIDP_PREPARSED_DATA * PHIDP_PREPARSED_DATA;
typedef USHORT USAGE, *PUSAGE;

typedef struct _HIDP_CAPS {
  USAGE  Usage;
  USAGE  UsagePage;
  USHORT  InputReportByteLength;
  USHORT  OutputReportByteLength;
  USHORT  FeatureReportByteLength;
  USHORT  Reserved[17];
  USHORT  NumberLinkCollectionNodes;
  USHORT  NumberInputButtonCaps;
  USHORT  NumberInputValueCaps;
  USHORT  NumberInputDataIndices;
  USHORT  NumberOutputButtonCaps;
  USHORT  NumberOutputValueCaps;
  USHORT  NumberOutputDataIndices;
  USHORT  NumberFeatureButtonCaps;
  USHORT  NumberFeatureValueCaps;
  USHORT  NumberFeatureDataIndices;
} HIDP_CAPS, *PHIDP_CAPS;

static dword FWriteTimeout = DEFAULT_TX_TIMEOUT;
static dword FReadTimeout = DEFAULT_RX_TIMEOUT;

// Use a "semaphore" here to avoid conflicting queue reading/writing
static CRITICAL_SECTION FQueueLock;

typedef VOID (__stdcall *pHidD_GetHidGuid)(LPGUID);
typedef BOOLEAN (__stdcall *pHidD_GetAttributes)(HANDLE, PHIDD_ATTRIBUTES);
typedef BOOLEAN (__stdcall *pHidD_SetFeature)(HANDLE, PVOID, ULONG);
typedef BOOLEAN (__stdcall *pHidD_GetFeature)(HANDLE, PVOID, ULONG);
typedef BOOLEAN (__stdcall *pHidD_GetManufacturerString)(HANDLE, PVOID, ULONG);
typedef BOOLEAN (__stdcall *pHidD_GetProductString)(HANDLE, PVOID, ULONG);
typedef BOOLEAN (__stdcall *pHidD_GetSerialNumberString)(HANDLE, PVOID, ULONG);
typedef BOOLEAN (__stdcall *pHidD_SetNumInputBuffers)(HANDLE, ULONG);
typedef BOOLEAN (__stdcall *pHidD_GetPreparsedData) (HANDLE, PHIDP_PREPARSED_DATA *);
typedef BOOLEAN (__stdcall *pHidP_GetCaps) (HANDLE, PHIDP_CAPS);

static HINSTANCE hHID = NULL;
static HANDLE FDeviceHandle = INVALID_HANDLE_VALUE;
static HANDLE FReadDeviceHandle = INVALID_HANDLE_VALUE;
static pHidD_GetHidGuid HidD_GetHidGuid = NULL;
static pHidD_GetAttributes HidD_GetAttributes = NULL;
static pHidD_SetFeature HidD_SetFeature = NULL;
static pHidD_GetFeature HidD_GetFeature = NULL;
static pHidD_GetManufacturerString HidD_GetManufacturerString = NULL;
static pHidD_GetProductString HidD_GetProductString = NULL;
static pHidD_GetSerialNumberString HidD_GetSerialNumberString = NULL;
static pHidD_SetNumInputBuffers HidD_SetNumInputBuffers = NULL;
static pHidD_GetPreparsedData HidD_GetPreparsedData = NULL;
static pHidP_GetCaps HidP_GetCaps = NULL;

static USHORT InputReportByteLength;
static USHORT OutputReportByteLength;

#define USB_PARM_STR_LENGTS	128
static wchar_t manufacturer[USB_PARM_STR_LENGTS + 1] = L"";
static wchar_t product[USB_PARM_STR_LENGTS + 1] = L"";
static wchar_t serial[USB_PARM_STR_LENGTS + 1] = L"";

static TCHAR devicepath[MAX_PATH];

static HANDLE FStopEvent = NULL;
static HANDLE FInputThread = NULL;
static OVERLAPPED ol;

static bool FThreadRunning = false;

// CAVEAT: QUEUE_SIZE must be a power of 2 value
#define QUEUE_SIZE	256
#define QUEUE_MASK	(QUEUE_SIZE - 1)

typedef struct
{
	byte contents[QUEUE_SIZE];
	int head;
	int tail;
	int size;
} queue_t;

static queue_t rx_queue = { { 0 }, 0, 0, 0 };

static byte tx_buff[USB_MAX_OUT_REPORT_SIZE] = { 0 };		// including extra prefix report ID byte
static int tx_filled = 0;											// # of bytes in tx_buff

// Forward declarations
static void USB_ClearTXBuffer(void);

//---------------------------------------------------------------------------
bool USB_FillQueue(byte c)
{
	if (rx_queue.size >= QUEUE_SIZE)
		return false;

	rx_queue.contents[QUEUE_MASK & rx_queue.head] = c;
	rx_queue.head++;
	rx_queue.size++;

	return true;
}
//---------------------------------------------------------------------------
bool USB_ReadQueue(byte *c)
{
	if (!rx_queue.size)
		return false;

	EnterCriticalSection(&FQueueLock);
	/////////////////////////////////////////////////////
	*c = rx_queue.contents[QUEUE_MASK & rx_queue.tail];
	rx_queue.tail++;
	rx_queue.size--;
	/////////////////////////////////////////////////////
	LeaveCriticalSection(&FQueueLock);

	return true;
}
//---------------------------------------------------------------------------
void USB_UnLoadDLL(void)
{
	if (hHID)
		FreeLibrary(hHID);

	hHID = NULL;

	HidD_GetHidGuid = NULL;
	HidD_GetAttributes = NULL;
	HidD_SetFeature = NULL;
	HidD_GetFeature = NULL;
	HidD_GetManufacturerString = NULL;
	HidD_GetProductString = NULL;
	HidD_GetSerialNumberString = NULL;
	HidD_SetNumInputBuffers = NULL;
	HidD_GetPreparsedData = NULL;
	HidP_GetCaps = NULL;
}
//---------------------------------------------------------------------------
bool USB_LoadDLL(void)
{
	if (hHID)
		return true;	// Already loaded

	hHID = LoadLibrary("HID.DLL");

	if (!hHID)
		return Error_SetErr(USB_ERR_DLL_LOAD);

	HidD_GetHidGuid = (pHidD_GetHidGuid)
		GetProcAddress(hHID, "HidD_GetHidGuid");
	HidD_GetAttributes = (pHidD_GetAttributes)
		GetProcAddress(hHID, "HidD_GetAttributes");
	HidD_SetFeature = (pHidD_SetFeature)
		GetProcAddress(hHID, "HidD_SetFeature");
	HidD_GetFeature = (pHidD_GetFeature)
		GetProcAddress(hHID, "HidD_GetFeature");
	HidD_GetManufacturerString = (pHidD_GetManufacturerString)
		GetProcAddress(hHID, "HidD_GetManufacturerString");
	HidD_GetProductString = (pHidD_GetProductString)
		GetProcAddress(hHID, "HidD_GetProductString");
	HidD_GetSerialNumberString = (pHidD_GetSerialNumberString)
		GetProcAddress(hHID, "HidD_GetSerialNumberString");
	HidD_SetNumInputBuffers = (pHidD_SetNumInputBuffers)
		GetProcAddress(hHID, "HidD_SetNumInputBuffers");
	HidD_GetPreparsedData = (pHidD_GetPreparsedData)
		GetProcAddress(hHID, "HidD_GetPreparsedData");
	HidP_GetCaps = (pHidP_GetCaps)
		GetProcAddress(hHID, "HidP_GetCaps");

	if (!HidD_GetHidGuid || !HidD_GetAttributes || !HidD_SetFeature ||
		!HidD_GetFeature ||!HidD_GetManufacturerString ||
		!HidD_GetProductString || !HidD_GetSerialNumberString ||
		!HidD_SetNumInputBuffers || !HidD_GetPreparsedData || !HidP_GetCaps)
	{
		Error_SetErr(USB_ERR_DLL_ENTRY_POINT);
		USB_UnLoadDLL();

		return false;
	}

	return true;
}
//---------------------------------------------------------------------------
// returns true if an iLCD devices is found. If there are more than one
// iLCD devices found always the first one is used (simplified approach)
// The required device details are written to the static vars
// manufacturer, product, serial and devicepath
bool USB_FindDevices(void)
{
	// Get HID GUID
	GUID guid;
	SP_DEVICE_INTERFACE_DATA devData;
	HDEVINFO hDevInfo;
	DWORD dwIndex = 0;
	DWORD dwSize;
	PSP_DEVICE_INTERFACE_DETAIL_DATA pDetailData;

	// preset to not found
	*devicepath = '\0';
	Error_SetErr(0);

	if (!USB_LoadDLL())
		return false;

	HidD_GetHidGuid(&guid);

	// Open handle to the set of all HID devices
	hDevInfo = SetupDiGetClassDevs(
				  &guid, NULL, NULL, DIGCF_PRESENT | DIGCF_INTERFACEDEVICE);

	if (hDevInfo == INVALID_HANDLE_VALUE)
		return Error_SetErr(USB_ERR_ENUM_DEV_SET);

	devData.cbSize = sizeof(devData);

	// For each HID device...
	while (true)
	{
		if (!SetupDiEnumDeviceInterfaces(hDevInfo, 0, &guid, dwIndex, &devData))
		{
			int err = GetLastError();
			if (err != ERROR_NO_MORE_ITEMS)
				Error_SetErr(USB_ERR_ENUM);

			break;
		}

		dwIndex++;

		// Get device name
		SetupDiGetDeviceInterfaceDetail(hDevInfo, &devData, NULL, 0, &dwSize, NULL);

		pDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(dwSize);

		pDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

		if (SetupDiGetDeviceInterfaceDetail(hDevInfo, &devData, pDetailData,
											dwSize, &dwSize, NULL))
		{
			HANDLE hDevice;

			// Open device to handle
			hDevice = CreateFile(pDetailData->DevicePath,
					GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE,
					NULL, OPEN_EXISTING, 0, NULL);

			if (hDevice != INVALID_HANDLE_VALUE)
			{
				// Get vendor and product ID values
				HIDD_ATTRIBUTES attr;

				attr.Size = sizeof(attr);
				if (HidD_GetAttributes(hDevice, &attr))
				{
#if 0				// vendor and product id may change, so we take the product name
					if ((attr.VendorID == ILCD_VENDOR_ID)
						&& (attr.ProductID == ILCD_PRODUCT_ID))
#endif
					{
						HidD_GetManufacturerString(hDevice, manufacturer, 128);
						HidD_GetProductString(hDevice, product, 128);
						HidD_GetSerialNumberString(hDevice, serial, 128);

						if (wcscmp(product, L"iLCD Controller") == 0)
						{
							strcpy(devicepath, pDetailData->DevicePath);
						}
					}
				}

				CloseHandle(hDevice);
			}
		}

		free(pDetailData);

		if (*devicepath)
			break;			// found an iLCD device
	}

	SetupDiDestroyDeviceInfoList(hDevInfo);

	if (*devicepath == '\0' && !Error_GetLastErrNo())
	{
		Error_SetErr(USB_ERR_NO_DEVICES_FOUND);
		return false;
	}

	return true;
}
//---------------------------------------------------------------------------
static bool Exit(void)
{
	FThreadRunning = false;
	return 0;
}
//---------------------------------------------------------------------------
// Windows does not allow to read from an USB HID device without using an
// separate thread, as ReadFile is a blocking call and there is no
// timeout possibitity and no peek function working with USB HID devices.
// Let's praise Windows once more ....
static unsigned __stdcall USB_ReadThreadProc(void *arg)
{
	DWORD cbRet;
	char in_report[USB_MAX_IN_REPORT_SIZE * USB_NUM_BUFFERS];
	bool bRet;
	int count;
	int i;

	// Open another handle to the HID device for read

	FReadDeviceHandle = CreateFile(devicepath,
									GENERIC_READ | GENERIC_WRITE,
									FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
									FILE_FLAG_OVERLAPPED, NULL);

	if (FReadDeviceHandle == INVALID_HANDLE_VALUE)
	{
		Error_SetErr(USB_RD_CREATE_FILE_ERR);
		return Exit();
	}

	// Set # of buffers to speed up processing of input reports
	if (!HidD_SetNumInputBuffers(FReadDeviceHandle, USB_NUM_BUFFERS))
	{
		Error_SetErr(USB_RD_CREATE_FILE_ERR);
		return Exit();
	}

	memset(&ol, 0, sizeof(ol));
	ol.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	// Caveat: We read multiple buffers at once to avoid overflow on slow systems.
	// If we have e.g. a InputReportByteLength of 65, the first byte of every buffer
	// will be the report ID (added by windows) and the second byte of every buffer
	// flags the size used in this buffer. We have to process all buffers we get
	// in one chunk buffer by buffer in this way.

	while (true)
	{
		ResetEvent(ol.hEvent);
		// Read input report from device
		bRet = ReadFile(FReadDeviceHandle, in_report,
					InputReportByteLength * USB_NUM_BUFFERS, &cbRet, &ol);

		if (!bRet)
		{
			if (GetLastError() == ERROR_IO_PENDING)
			{
				// OK - Wait for data to come in or stop event to be set
				HANDLE handles[2];
				DWORD waitRet;

				handles[0] = ol.hEvent;
				handles[1] = FStopEvent;

				waitRet = WaitForMultipleObjects(2, handles, FALSE, INFINITE);

				if (waitRet == (WAIT_OBJECT_0 + 1))
				{
					// Stop event was set
					CloseHandle(FReadDeviceHandle);
					CloseHandle(ol.hEvent);
					DeleteCriticalSection(&FQueueLock);
					return Exit();
				}
				else if (waitRet == WAIT_OBJECT_0)
				{
					// Data came in
					bRet = GetOverlappedResult(FReadDeviceHandle, &ol, &cbRet, TRUE);
				}
			}
		}

		if (!bRet)
		{
			Error_SetErr(USB_ERR_CONNECTION_LOST);

			CloseHandle(FReadDeviceHandle);
			CloseHandle(ol.hEvent);
			return Exit();
		}
		else
		{
			// We write any data we get into the rx_queue.
			// The routine called from main() retrieves data from this queue then.

			// cbRet is the # of bytes we really got, this will be most of the time
			// InputReportByteLength, but it may be a multiple of InputReportByteLength
			// as well.

			EnterCriticalSection(&FQueueLock);
			/////////////////////////////////////////////////////
			count = 0;
			for (i = 0; i < (int) cbRet; i++)
			{
				if ((i % InputReportByteLength) == 0)
				{
					// reset count
					count = 0;
				}
				else if ((i % InputReportByteLength) == 1)
				{
					// # of bytes used in this buffer
					count = in_report[i];
				}
				else if (count)
				{
					// get "real" data
					USB_FillQueue(in_report[i]);
					count--;
				}
			}
			/////////////////////////////////////////////////////
			LeaveCriticalSection(&FQueueLock);
		}
	}
}
//---------------------------------------------------------------------------
static HANDLE USB_CreateInputThread(void)
{
	unsigned dwThreadID;
	HANDLE hThread;
	SECURITY_ATTRIBUTES sa =
	{
		sizeof(SECURITY_ATTRIBUTES), /* structure size */
		0,      /* No security descriptor */
		true    /* Thread handle is inheritable */
	};

	FStopEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	// set the state of the specified event object to nonsignaled
	ResetEvent(FStopEvent);

	hThread = (HANDLE)_beginthreadex(
			&sa,          /* Thread security */
			4096,         /* Thread stack size */
			USB_ReadThreadProc,   /* Thread starting address */
			(void *)0,    /* Thread start argument */
			CREATE_SUSPENDED,  /* Create in suspended state */
			&dwThreadID);   /* Thread ID */

	if (!hThread)
	{
		Error_SetErr(USB_RD_THREAD_CREATE_ERROR);
		CloseHandle(FStopEvent);
		FStopEvent = NULL;
	}

	InitializeCriticalSection(&FQueueLock);

	return hThread;
}
//---------------------------------------------------------------------------
bool USB_ReceiveChar(unsigned char *c)
{
	DWORD timestamp = GetTickCount() + FReadTimeout;
	bool rv = false;

	while (GetTickCount() < timestamp)
	{
		if (USB_ReadQueue(c))
		{
			rv = true;
			break;
		}
		// give the receive thread more chance to run
		Sleep(0);
	}

	if (!rv)
		return Error_SetErr(USB_ERR_RX_NO_DATA);

	return true;
}
//---------------------------------------------------------------------------
bool USB_Open(void)
{
	int i;
	void *data;
	bool err = false;
	PHIDP_PREPARSED_DATA	PreparsedData;
	HIDP_CAPS Capabilities;

	if (FDeviceHandle != INVALID_HANDLE_VALUE)
		return Error_SetErr(USB_ERR_ALREADY_OPEN);

	if (!USB_FindDevices())
		return false;

	// Open device to handle
	FDeviceHandle = CreateFile(devicepath,
			GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, 0, NULL);

	if (FDeviceHandle == INVALID_HANDLE_VALUE)
		return Error_SetErr(USB_ERR_WR_CREATE_FILE);

	if (HidD_GetPreparsedData(FDeviceHandle, &PreparsedData))
	{
		err = (HidP_GetCaps(PreparsedData, &Capabilities) != HIDP_STATUS_SUCCESS);
	}
	else
		err = true;

	if (!err)
	{
		// Set # of buffers to speed up processing of input reports
		err = !HidD_SetNumInputBuffers(FDeviceHandle, USB_NUM_BUFFERS);
	}

	if (err)
	{
		CloseHandle(FDeviceHandle);
		FDeviceHandle = INVALID_HANDLE_VALUE;

		return Error_SetErr(USB_ERR_WR_CREATE_FILE);
	}

	// Get size of input and output reports
	InputReportByteLength = Capabilities.InputReportByteLength;
	OutputReportByteLength = Capabilities.OutputReportByteLength;

	if (InputReportByteLength > USB_MAX_IN_REPORT_SIZE ||
		OutputReportByteLength > USB_MAX_OUT_REPORT_SIZE)
	{
		// Should never happen!

		CloseHandle(FDeviceHandle);
		FDeviceHandle = INVALID_HANDLE_VALUE;

		return Error_SetErr(USB_ERR_WR_CREATE_FILE);
	}

	if ((FInputThread = USB_CreateInputThread()) == NULL)
	{
		USB_Close();
		return false;
	}
	ResumeThread(FInputThread);
	FThreadRunning = true;

	USB_ClearTXBuffer();

	return true;
}
//---------------------------------------------------------------------------
void USB_SetTimeouts(dword rx_tout, dword tx_tout)
{
	FReadTimeout = rx_tout;
	FWriteTimeout = tx_tout;
}
//---------------------------------------------------------------------------
bool USB_Close(void)
{
	if (FDeviceHandle == INVALID_HANDLE_VALUE)
		return Error_SetErr(USB_ERR_NOT_OPEN);

	if (FStopEvent)
		SetEvent(FStopEvent);

	if (FInputThread)
	{
		// close the input thread's handle
		CloseHandle(FInputThread);
		DeleteCriticalSection(&FQueueLock);
		FInputThread = NULL;
	}

	CloseHandle(FDeviceHandle);
	FDeviceHandle = INVALID_HANDLE_VALUE;

	while (FThreadRunning)
		Sleep(100);

	if (FStopEvent)
	{
		CloseHandle(FStopEvent);
		FStopEvent = NULL;
	}

	USB_UnLoadDLL();
	return true;
}
//---------------------------------------------------------------------------
bool USB_IsDevOpen(void)
{
	return FDeviceHandle != INVALID_HANDLE_VALUE;
}
//---------------------------------------------------------------------------
bool USB_Transmit(byte *data, DWORD lenght)
{
	return USB_TransmitWTimeout(data, lenght, FWriteTimeout);
}
//---------------------------------------------------------------------------
static void USB_ClearTXBuffer(void)
{
	// clear tx buffer
	tx_filled = 0;
	memset(tx_buff, 0, sizeof(tx_buff));
}
//---------------------------------------------------------------------------
static bool USB_FlushTXData(word timeout)
{
	DWORD timestamp;
	DWORD written;
	int last_err = 0;
	bool rv = true;

	if (tx_filled)
	{
		timestamp = GetTickCount();

		while (GetTickCount() < timestamp + (int) timeout)
		{
			// write OutputReportByteLength bytes
			rv = WriteFile(FDeviceHandle, tx_buff, OutputReportByteLength, &written, NULL);

			if (rv)
				break;

			// Note: The iLCD device disables write when it is busy.
			// Depending on the PC's operating system different errors may be
			// reported by GetLastError() below, so we only stop when we either
			// get a device not existent error or we run into our timeout.

			last_err = GetLastError();

			if (last_err == ERROR_DEV_NOT_EXIST)
				break;
		}

		if (!rv)
		{
			if (last_err != ERROR_DEV_NOT_EXIST)
				Error_SetErr(USB_ERR_WRITE_TIMEOUT);
			else
				Error_SetErr(USB_ERR_WRITE);
		}
		else if ((int) written != OutputReportByteLength)
		{
			Error_SetErr(USB_ERR_PARTIAL_WRITE);

			rv = false;
		}
	}

	// clear TX buffer now
	USB_ClearTXBuffer();

	return rv;
}
//---------------------------------------------------------------------------
static bool USB_FillData(byte *data, DWORD lenght, word timeout)
{
	DWORD i;

	for (i = 0; i < lenght; i++)
	{
		if (tx_filled >= OutputReportByteLength - 2)
		{
			if (!USB_FlushTXData(timeout))
				return false;
		}

		tx_buff[tx_filled + 2] = data[i];
		tx_buff[0] = 0;
		tx_buff[1] = ++tx_filled;					   // index 1 contains # of bytes
	}

	return true;
}
//---------------------------------------------------------------------------
bool USB_TransmitWTimeout(byte *data, DWORD lenght, word timeout)
{
	if (!USB_IsDevOpen())
		return Error_SetErr(USB_ERR_NOT_OPEN);

	Error_SetErr(0);

	if (!USB_FillData(data, lenght, timeout))
	{
		USB_ClearTXBuffer();
		return false;
	}

	return USB_FlushTXData(timeout);
}
//---------------------------------------------------------------------------

