//---------------------------------------------------------------------------
#ifndef __BORLANDC__
	// VC++ needs to know that it must link setupapi.lib
	#pragma comment(lib, "setupapi.lib")
	//  Use the -tWM  (32-bit multi-threaded target) command-line switch
	#pragma checkoption -tWM
#else
	#if __BORLANDC__ >= 0x0630
		#define CPP_BUILDER_XE		1
	#endif
	#ifndef CPP_BUILDER_XE
		#define __in
		#define __out
	#endif
#endif

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <setupapi.h>
#include <process.h>

#include "winusb.h"
#include "commands.h"
#include "errmsgs.h"

// ********** Stuff needed to talk to Windows DLL "WinUSB.dll" **********

// Pipe policy types
#define SHORT_PACKET_TERMINATE  0x01
#define AUTO_CLEAR_STALL        0x02
#define PIPE_TRANSFER_TIMEOUT   0x03
#define IGNORE_SHORT_PACKETS    0x04
#define ALLOW_PARTIAL_READS     0x05
#define AUTO_FLUSH              0x06
#define RAW_IO                  0x07
#define MAXIMUM_TRANSFER_SIZE   0x08

// Power policy types
//
// Add 0x80 for Power policy types in order to prevent overlap with
// Pipe policy types to prevent "accidentally" setting the wrong value for the
// wrong type.
//
#define AUTO_SUSPEND            0x81
#define SUSPEND_DELAY           0x83

// Device Information types
#define DEVICE_SPEED            0x01

// Device Speeds
#define LowSpeed                0x01
#define FullSpeed               0x02
#define HighSpeed               0x03

//
// Endpoint direction bit, stored in address
//

#define USB_ENDPOINT_DIRECTION_MASK      0x80

// test direction bit in the bEndpointAddress field of
// an endpoint descriptor.
#define USB_ENDPOINT_DIRECTION_OUT(addr) (!((addr) & USB_ENDPOINT_DIRECTION_MASK))
#define USB_ENDPOINT_DIRECTION_IN(addr)  ((addr) & USB_ENDPOINT_DIRECTION_MASK)
#define USB_STRING_DESCRIPTOR_TYPE	3

// {42CA71EC-CE1C-44C2-82DE-87D8D8FF6C1E}
#include <initguid.h>

// {ED48CB1D-D414-4427-B35F-5109F4305785}
static const GUID WINUSB_DEVICE_INTERFACE =
{ 0xed48cb1d, 0xd414, 0x4427, { 0xb3, 0x5f, 0x51, 0x9, 0xf4, 0x30, 0x57, 0x85 } };

	typedef PVOID WINUSB_INTERFACE_HANDLE, *PWINUSB_INTERFACE_HANDLE;

#pragma option -a1

	typedef struct _WINUSB_SETUP_PACKET {
		 UCHAR   RequestType;
		 UCHAR   Request;
		 USHORT  Value;
		 USHORT  Index;
		 USHORT  Length;
	} WINUSB_SETUP_PACKET, *PWINUSB_SETUP_PACKET;

#pragma option -a

	// ******** CAVEAT ********

	// enums must be defined as int's (size = 4 bytes) otherwise
	// _WINUSB_PIPE_INFORMATION below has wrong size !

	// Ensure your compiler has set the corresponding option to
	// use enum's as int's !

	typedef enum _USBD_PIPE_TYPE {
	  UsbdPipeTypeControl       = 0,
	  UsbdPipeTypeIsochronous   = 1,
	  UsbdPipeTypeBulk          = 2,
	  UsbdPipeTypeInterrupt     = 3
	} USBD_PIPE_TYPE;

	typedef struct _WINUSB_PIPE_INFORMATION {
		 USBD_PIPE_TYPE  PipeType;
		 UCHAR           PipeId;
		 USHORT          MaximumPacketSize;
		 UCHAR           Interval;
	} WINUSB_PIPE_INFORMATION, *PWINUSB_PIPE_INFORMATION;

	typedef struct _USB_INTERFACE_DESCRIPTOR {
	  UCHAR bLength;
	  UCHAR bDescriptorType;
	  UCHAR bInterfaceNumber;
	  UCHAR bAlternateSetting;
	  UCHAR bNumEndpoints;
	  UCHAR bInterfaceClass;
	  UCHAR bInterfaceSubClass;
	  UCHAR bInterfaceProtocol;
	  UCHAR iInterface;
	} USB_INTERFACE_DESCRIPTOR, *PUSB_INTERFACE_DESCRIPTOR;

	typedef struct _USB_CONFIGURATION_DESCRIPTOR {
	  UCHAR  bLength;
	  UCHAR  bDescriptorType;
	  USHORT wTotalLength;
	  UCHAR  bNumInterfaces;
	  UCHAR  bConfigurationValue;
	  UCHAR  iConfiguration;
	  UCHAR  bmAttributes;
	  UCHAR  MaxPower;
	} USB_CONFIGURATION_DESCRIPTOR, *PUSB_CONFIGURATION_DESCRIPTOR;

	typedef struct _USB_COMMON_DESCRIPTOR {
	  UCHAR bLength;
	  UCHAR bDescriptorType;
	} USB_COMMON_DESCRIPTOR, *PUSB_COMMON_DESCRIPTOR;

	typedef BOOL (__stdcall	*pWinUsb_Initialize)(
		 __in HANDLE DeviceHandle,
		 __out PWINUSB_INTERFACE_HANDLE InterfaceHandle
		 );

	typedef BOOL (__stdcall *pWinUsb_Free)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle
		 );

	typedef BOOL (__stdcall *pWinUsb_GetAssociatedInterface)(
		 __in WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in UCHAR AssociatedInterfaceIndex,
		 __out PWINUSB_INTERFACE_HANDLE AssociatedInterfaceHandle
		 );

	typedef BOOL (__stdcall *pWinUsb_GetDescriptor)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR DescriptorType,
		 __in  UCHAR Index,
		 __in  USHORT LanguageID,
		 __out PUCHAR Buffer,
		 __in  ULONG BufferLength,
		 __out PULONG LengthTransferred
		 );

	typedef BOOL (__stdcall *pWinUsb_QueryInterfaceSettings)(
		 __in WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in UCHAR AlternateInterfaceNumber,
		 __out PUSB_INTERFACE_DESCRIPTOR UsbAltInterfaceDescriptor
		 );

	typedef BOOL (__stdcall *pWinUsb_QueryDeviceInformation)(
		 __in WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in ULONG InformationType,
		 __in __out PULONG BufferLength,
		 __out PVOID Buffer
		 );

	typedef BOOL (__stdcall *pWinUsb_SetCurrentAlternateSetting)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR SettingNumber
		 );

	typedef BOOL (__stdcall *pWinUsb_GetCurrentAlternateSetting)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __out PUCHAR SettingNumber
		 );

	typedef BOOL (__stdcall *pWinUsb_QueryPipe)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR AlternateInterfaceNumber,
		 __in  UCHAR PipeIndex,
		 __out PWINUSB_PIPE_INFORMATION PipeInformation
		 );

	typedef BOOL (__stdcall *pWinUsb_SetPipePolicy)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR PipeID,
		 __in  ULONG PolicyType,
		 __in  ULONG ValueLength,
		 __in  PVOID Value
		 );

	typedef BOOL (__stdcall *pWinUsb_GetPipePolicy)(
		 __in WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in UCHAR PipeID,
		 __in ULONG PolicyType,
		 __in __out PULONG ValueLength,
		 __out PVOID Value
		 );

	typedef BOOL (__stdcall *pWinUsb_ReadPipe)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR PipeID,
		 __in  PUCHAR Buffer,
		 __in  ULONG BufferLength,
		 __out PULONG LengthTransferred,
		 __in  LPOVERLAPPED Overlapped
		 );

	typedef BOOL (__stdcall *pWinUsb_WritePipe)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR PipeID,
		 __in  PUCHAR Buffer,
		 __in  ULONG BufferLength,
		 __out PULONG LengthTransferred,
		 __in  LPOVERLAPPED Overlapped
		 );

	typedef BOOL (__stdcall *pWinUsb_ControlTransfer)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  WINUSB_SETUP_PACKET SetupPacket,
		 __in  PUCHAR Buffer,
		 __in  ULONG BufferLength,
		 __out PULONG LengthTransferred,
		 __in  LPOVERLAPPED Overlapped
		 );

	typedef BOOL (__stdcall *pWinUsb_ResetPipe)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR PipeID
		 );

	typedef BOOL (__stdcall *pWinUsb_AbortPipe)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR PipeID
		 );

	typedef BOOL (__stdcall *pWinUsb_FlushPipe)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  UCHAR PipeID
		 );

	typedef BOOL (__stdcall *pWinUsb_SetPowerPolicy)(
		 __in  WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in  ULONG PolicyType,
		 __in  ULONG ValueLength,
		 __in  PVOID Value
		 );

	typedef BOOL (__stdcall *pWinUsb_GetPowerPolicy)(
		 __in WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in ULONG PolicyType,
		 __in __out PULONG ValueLength,
		 __out PVOID Value
		 );

	typedef BOOL (__stdcall *pWinUsb_GetOverlappedResult)(
		 __in WINUSB_INTERFACE_HANDLE InterfaceHandle,
		 __in LPOVERLAPPED lpOverlapped,
		 __out LPDWORD lpNumberOfBytesTransferred,
		 BOOL bWait
		 );

	typedef PUSB_INTERFACE_DESCRIPTOR (__stdcall *pWinUsb_ParseConfigurationDescriptor)(
		 __in PUSB_CONFIGURATION_DESCRIPTOR ConfigurationDescriptor,
		 __in PVOID StartPosition,
		 __in LONG InterfaceNumber,
		 __in LONG AlternateSetting,
		 __in LONG InterfaceClass,
		 __in LONG InterfaceSubClass,
		 __in LONG InterfaceProtocol
		 );

	typedef PUSB_COMMON_DESCRIPTOR (__stdcall *pWinUsb_ParseDescriptors)(
		 __in PVOID    DescriptorBuffer,
		 __in ULONG    TotalLength,
		 __in PVOID    StartPosition,
		 __in LONG     DescriptorType
		 );

	typedef struct
	{
		UCHAR PipeInId;
		UCHAR PipeOutId;
	} PIPE_ID;

	// define function pointers to DLL funcs
	#define FUNC(x)	p##x x;
		  

	FUNC(WinUsb_Initialize);
	FUNC(WinUsb_Free);
	FUNC(WinUsb_GetAssociatedInterface);
	FUNC(WinUsb_GetDescriptor);
	FUNC(WinUsb_QueryInterfaceSettings);
	FUNC(WinUsb_QueryDeviceInformation);
	FUNC(WinUsb_SetCurrentAlternateSetting);
	FUNC(WinUsb_GetCurrentAlternateSetting);
	FUNC(WinUsb_QueryPipe);
	FUNC(WinUsb_SetPipePolicy);
	FUNC(WinUsb_GetPipePolicy);
	FUNC(WinUsb_ReadPipe);
	FUNC(WinUsb_WritePipe);
	FUNC(WinUsb_ControlTransfer);
	FUNC(WinUsb_ResetPipe);
	FUNC(WinUsb_AbortPipe);
	FUNC(WinUsb_FlushPipe);
	FUNC(WinUsb_SetPowerPolicy);
	FUNC(WinUsb_GetPowerPolicy);
	FUNC(WinUsb_GetOverlappedResult);
	FUNC(WinUsb_ParseConfigurationDescriptor);
	FUNC(WinUsb_ParseDescriptors);

	#undef FUNC

	PIPE_ID FPipeID;

// ********** End of "WinUSB.dll" stuff **********

static int FWriteTimeout = DEFAULT_TX_TIMEOUT;
static int FCurrWriteTimeout = 0;
static int FReadTimeout = DEFAULT_RX_TIMEOUT;

static int FReadPacketSize;
static int FWritePacketSize;

// Use a "semaphore" here to avoid conflicting queue reading/writing
static CRITICAL_SECTION FQueueLock;

static HINSTANCE hWinUSB = NULL;
static HANDLE FDeviceHandle = INVALID_HANDLE_VALUE;
static WINUSB_INTERFACE_HANDLE FWinUSBHandle = INVALID_HANDLE_VALUE;

#define USB_PARM_STR_LENGTS	128
static wchar_t manufacturer[USB_PARM_STR_LENGTS + 1] = L"";
static wchar_t product[USB_PARM_STR_LENGTS + 1] = L"";
static wchar_t serial[USB_PARM_STR_LENGTS + 1] = L"";

static TCHAR devicepath[MAX_PATH] = "";

static HANDLE  FStopEvent = NULL;
static HANDLE FInputThread = NULL;
static OVERLAPPED ol;

static bool FThreadRunning = false;

// CAVEAT: QUEUE_SIZE must be a power of 2 value
#define QUEUE_SIZE	256
#define QUEUE_MASK	(QUEUE_SIZE - 1)

typedef struct
{
	byte contents[QUEUE_SIZE];
	int head;
	int tail;
	int size;
} queue_t;

static queue_t rx_queue = { { 0 }, 0, 0, 0 };

typedef struct usb_device_descriptor_t
{
	byte bLength;
	byte bDescriptorType;
	ushort bcdUSB;
	byte bDeviceClass;
	byte bDeviceSubClass;
	byte bDeviceProtocol;
	byte bMaxPacketSize0;
	ushort idVendor;
	ushort idProduct;
	ushort bcdDevice;
	byte iManufacturer;
	byte iProduct;
	byte iSerialNumber;
	byte bNumConfigurations;
}usb_device_descriptor_t;

#define USB_DEVICE_DESCRIPTOR_TYPE 0x01

// forward declaration
static bool WinUSB_CloseDeviceHandles(void);

//---------------------------------------------------------------------------
void WinUSB_EmptyQueue(void)
{
	rx_queue.head = 0;
	rx_queue.tail = 0;
	rx_queue.size = 0;
}
//---------------------------------------------------------------------------
bool WinUSB_FillQueue(byte c)
{
	if (rx_queue.size >= QUEUE_SIZE)
		return false;

	rx_queue.contents[QUEUE_MASK & rx_queue.head] = c;
	rx_queue.head++;
	rx_queue.size++;

	return true;
}
//---------------------------------------------------------------------------
bool WinUSB_ReadQueue(byte *c)
{
	if (!rx_queue.size)
		return false;

	EnterCriticalSection(&FQueueLock);
	/////////////////////////////////////////////////////
	*c = rx_queue.contents[QUEUE_MASK & rx_queue.tail];
	rx_queue.tail++;
	rx_queue.size--;
	/////////////////////////////////////////////////////
	LeaveCriticalSection(&FQueueLock);

	return true;
}
//---------------------------------------------------------------------------
void WinUSB_UnLoadDLL(void)
{
	#define SET_NULL(x)	x = NULL

	if (!hWinUSB)
		return;

	FreeLibrary(hWinUSB);
	hWinUSB = NULL;

	SET_NULL(WinUsb_Initialize);
	SET_NULL(WinUsb_Free);
	SET_NULL(WinUsb_GetAssociatedInterface);
	SET_NULL(WinUsb_GetDescriptor);
	SET_NULL(WinUsb_QueryInterfaceSettings);
	SET_NULL(WinUsb_QueryDeviceInformation);
	SET_NULL(WinUsb_SetCurrentAlternateSetting);
	SET_NULL(WinUsb_GetCurrentAlternateSetting);
	SET_NULL(WinUsb_QueryPipe);
	SET_NULL(WinUsb_SetPipePolicy);
	SET_NULL(WinUsb_GetPipePolicy);
	SET_NULL(WinUsb_ReadPipe);
	SET_NULL(WinUsb_WritePipe);
	SET_NULL(WinUsb_ControlTransfer);
	SET_NULL(WinUsb_ResetPipe);
	SET_NULL(WinUsb_AbortPipe);
	SET_NULL(WinUsb_FlushPipe);
	SET_NULL(WinUsb_SetPowerPolicy);
	SET_NULL(WinUsb_GetPowerPolicy);
	SET_NULL(WinUsb_GetOverlappedResult);
	SET_NULL(WinUsb_ParseConfigurationDescriptor);
	SET_NULL(WinUsb_ParseDescriptors);

	#undef SET_NULL
}
//---------------------------------------------------------------------------
static bool WinUSB_ExitWithErr(void)
{
	Error_SetErr(USB_ERR_DLL_ENTRY_POINT);
	WinUSB_UnLoadDLL();

	return false;
}
//---------------------------------------------------------------------------
bool WinUSB_LoadDLL(void)
{
	#define GET_PROC_ADDR(x)	x = (p##x) GetProcAddress(hWinUSB, #x);	\
										if (x == NULL) return WinUSB_ExitWithErr()

	if (hWinUSB)
		return true;	// Already loaded

	hWinUSB = LoadLibrary("WinUSB.DLL");

	if (!hWinUSB)
		return Error_SetErr(USB_ERR_DLL_LOAD);

	GET_PROC_ADDR(WinUsb_Initialize);
	GET_PROC_ADDR(WinUsb_Free);
	GET_PROC_ADDR(WinUsb_GetAssociatedInterface);
	GET_PROC_ADDR(WinUsb_GetDescriptor);
	GET_PROC_ADDR(WinUsb_QueryInterfaceSettings);
	GET_PROC_ADDR(WinUsb_QueryDeviceInformation);
	GET_PROC_ADDR(WinUsb_SetCurrentAlternateSetting);
	GET_PROC_ADDR(WinUsb_GetCurrentAlternateSetting);
	GET_PROC_ADDR(WinUsb_QueryPipe);
	GET_PROC_ADDR(WinUsb_SetPipePolicy);
	GET_PROC_ADDR(WinUsb_GetPipePolicy);
	GET_PROC_ADDR(WinUsb_ReadPipe);
	GET_PROC_ADDR(WinUsb_WritePipe);
	GET_PROC_ADDR(WinUsb_ControlTransfer);
	GET_PROC_ADDR(WinUsb_ResetPipe);
	GET_PROC_ADDR(WinUsb_AbortPipe);
	GET_PROC_ADDR(WinUsb_FlushPipe);
	GET_PROC_ADDR(WinUsb_SetPowerPolicy);
	GET_PROC_ADDR(WinUsb_GetPowerPolicy);
	GET_PROC_ADDR(WinUsb_GetOverlappedResult);
	GET_PROC_ADDR(WinUsb_ParseConfigurationDescriptor);
	GET_PROC_ADDR(WinUsb_ParseDescriptors);

	return true;
}
//---------------------------------------------------------------------------
// Returns true if an iLCD device is found. If there are more than one
// iLCD devices found always the first one is used (simplified approach).
// The required device detail is written to devicepath.
bool WinUSB_FindDevices(void)
{
	GUID guid = WINUSB_DEVICE_INTERFACE;
	SP_DEVICE_INTERFACE_DATA devData;
	PSP_DEVICE_INTERFACE_DETAIL_DATA pDetailData;
	DWORD dwSize;
	HDEVINFO hDevInfo;
	DWORD dwIndex = 0;

	// preset to not found
	*devicepath = '\0';
	Error_SetErr(0);
	WinUSB_EmptyQueue();

	if (!WinUSB_LoadDLL())
		return false;

	// device interface class.
	hDevInfo = SetupDiGetClassDevs(&guid, NULL, NULL, DIGCF_PRESENT | DIGCF_INTERFACEDEVICE);

	if (hDevInfo == INVALID_HANDLE_VALUE)
		return -1;

	devData.cbSize = sizeof(devData);

	// For each device...
	while (true)
	{
		if (!SetupDiEnumDeviceInterfaces(hDevInfo, 0, &guid, dwIndex, &devData))
		{
			int err = GetLastError();

			if (err != ERROR_NO_MORE_ITEMS)
				Error_SetErr(USB_ERR_ENUM);

			break;
		}

		dwIndex++;

		// Get device name
		SetupDiGetDeviceInterfaceDetail(hDevInfo, &devData, NULL, 0, &dwSize, NULL);

		pDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(dwSize);

		pDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

		if (SetupDiGetDeviceInterfaceDetail(hDevInfo, &devData, pDetailData, dwSize, &dwSize, NULL))
		{
			// get device handle
			HANDLE hDevice = CreateFile(	pDetailData->DevicePath,
													GENERIC_READ | GENERIC_WRITE,
													FILE_SHARE_READ | FILE_SHARE_WRITE,
													NULL, OPEN_EXISTING,
													FILE_FLAG_OVERLAPPED,
													NULL);

			if (hDevice != INVALID_HANDLE_VALUE)
			{

				WINUSB_INTERFACE_HANDLE hWinUSBHandle = INVALID_HANDLE_VALUE;
				unsigned long LengthTransferred;

				if (WinUsb_Initialize(hDevice, &hWinUSBHandle))
				{
					usb_device_descriptor_t device_descriptor;

					if (WinUsb_GetDescriptor(hWinUSBHandle, 1 , 0, 0,
						(PUCHAR)&device_descriptor, sizeof(device_descriptor), &LengthTransferred))
					{
						if (LengthTransferred >= 4 &&
							 (device_descriptor.idProduct == 0x1511 ||
							  device_descriptor.idProduct == 0x1512 ||
							  device_descriptor.idProduct == 0x1513)&&
							 device_descriptor.idVendor == 0x0483)
						{
							strcpy(devicepath, pDetailData->DevicePath);
						}
					}
				}

				CloseHandle(hDevice);

				if (hWinUSBHandle != INVALID_HANDLE_VALUE)
					WinUsb_Free(hWinUSBHandle);
			}
		}

		free(pDetailData);

		if (*devicepath)
			break;			// found an iLCD device
	}

	SetupDiDestroyDeviceInfoList(hDevInfo);

	if (*devicepath == '\0' && !Error_GetLastErrNo())
	{
		Error_SetErr(USB_ERR_NO_DEVICES_FOUND);
		return false;
	}

	return true;
}
//---------------------------------------------------------------------------
static bool Exit(void)
{
	FThreadRunning = false;
	return 0;
}
//---------------------------------------------------------------------------
static unsigned __stdcall WinUSB_ReadThreadProc(void *arg)
{
	#define BUFF_COUNT 32

	int read_size = FReadPacketSize * BUFF_COUNT;
	DWORD cbRead;
	BOOL bRet;
	byte *buff;
	int i;

	buff = malloc(read_size);

	if (buff == NULL) 		// should never happen
		return Exit();

	memset(&ol, 0, sizeof(ol));
	ol.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	while (true)
	{
		ResetEvent(ol.hEvent);

		bRet = WinUsb_ReadPipe(FWinUSBHandle, FPipeID.PipeInId, buff, read_size, &cbRead, &ol);

		if (!bRet)
		{
			if (GetLastError() == ERROR_IO_PENDING)
			{
				// OK - Wait for data to come in or stop event to be set
				HANDLE handles[2];
				DWORD waitRet;

				handles[0] = ol.hEvent;
				handles[1] = FStopEvent;

				waitRet = WaitForMultipleObjects(2, handles, FALSE, INFINITE);

				if (waitRet == (WAIT_OBJECT_0 + 1))
				{
					// Stop event was set
					CloseHandle(ol.hEvent);

					free(buff);
					return Exit();
				}
				else if (waitRet == WAIT_OBJECT_0)
				{
					// Data came in
					bRet = WinUsb_GetOverlappedResult(FWinUSBHandle, &ol, &cbRead, TRUE);
				}
			}
		}

		if (!bRet)
		{
			Error_SetErr(USB_ERR_CONNECTION_LOST);

			CloseHandle(ol.hEvent);
			WinUSB_CloseDeviceHandles();
			DeleteCriticalSection(&FQueueLock);

			return Exit();
		}
		else
		{
			// We write any data we get into the rx_queue.
			// The routine called from main() retrieves data from this queue then.

			EnterCriticalSection(&FQueueLock);
			/////////////////////////////////////////////////////
			for (i = 0; i < (int) cbRead; i++)
			{
				WinUSB_FillQueue(buff[i]);
			}
			/////////////////////////////////////////////////////
			LeaveCriticalSection(&FQueueLock);

		}
	}
}
//---------------------------------------------------------------------------
static HANDLE WinUSB_CreateInputThread(void)
{
	unsigned dwThreadID;
	HANDLE hThread;
	SECURITY_ATTRIBUTES sa =
	{
		sizeof(SECURITY_ATTRIBUTES), /* structure size */
		0,      /* No security descriptor */
		true    /* Thread handle is inheritable */
	};

	FStopEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	// set the state of the specified event object to nonsignaled
	ResetEvent(FStopEvent);

	hThread = (HANDLE)_beginthreadex(
			&sa,          /* Thread security */
			4096,         /* Thread stack size */
			WinUSB_ReadThreadProc,   /* Thread starting address */
			(void *)0,    /* Thread start argument */
			CREATE_SUSPENDED,  /* Create in suspended state */
			&dwThreadID);   /* Thread ID */

	if (!hThread)
	{
		Error_SetErr(USB_RD_THREAD_CREATE_ERROR);
		CloseHandle(FStopEvent);
		FStopEvent = NULL;
	}

	InitializeCriticalSection(&FQueueLock);

	return hThread;
}
//---------------------------------------------------------------------------
bool WinUSB_ReceiveChar(unsigned char *c)
{
	DWORD timestamp = GetTickCount() + FReadTimeout;
	bool rv = false;

	while (GetTickCount() < timestamp)
	{
		if (WinUSB_ReadQueue(c))
		{
			rv = true;
			break;
		}
		// give the receive thread more chance to run
		Sleep(0);
	}

	if (!rv)
		return Error_SetErr(USB_ERR_RX_NO_DATA);

	return true;
}
//---------------------------------------------------------------------------
static bool WinUSB_CloseDeviceHandles(void)
{
	bool rv = true;

	if (FWinUSBHandle != INVALID_HANDLE_VALUE) 
	{
		WinUsb_Free(FWinUSBHandle);
		FWinUSBHandle = INVALID_HANDLE_VALUE;
	}
	else
		rv = false;

	if (FDeviceHandle != INVALID_HANDLE_VALUE) 
	{
		CloseHandle(FDeviceHandle);
		FDeviceHandle = INVALID_HANDLE_VALUE;
	}
	else
		rv = false;

	FWriteTimeout = -1;
	return rv;
}
//---------------------------------------------------------------------------
bool WinUSB_Open(void)
{
	USB_INTERFACE_DESCRIPTOR InterfaceDescriptor;
	WINUSB_PIPE_INFORMATION  Pipe;
	int err, index;
	dword written, timeout = 100;
	byte dummy_buff = 0;
	
	if (FDeviceHandle != INVALID_HANDLE_VALUE)
		return Error_SetErr(USB_ERR_ALREADY_OPEN);

	if (!WinUSB_FindDevices())
		return false;

	// Open device to handle
	FDeviceHandle = CreateFile(devicepath,
			GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);

	if (FDeviceHandle == INVALID_HANDLE_VALUE)
		return Error_SetErr(USB_ERR_WR_CREATE_FILE);

	FWinUSBHandle = INVALID_HANDLE_VALUE;
		
	if (!WinUsb_Initialize(FDeviceHandle, &FWinUSBHandle))
	{
		CloseHandle(FDeviceHandle);
		FDeviceHandle = INVALID_HANDLE_VALUE;
		
		return Error_SetErr(USB_ERR_WR_CREATE_FILE);
	}  

	memset(&InterfaceDescriptor, 0, sizeof(USB_INTERFACE_DESCRIPTOR));
	memset(&Pipe, 0, sizeof(WINUSB_PIPE_INFORMATION));

	FPipeID.PipeInId = 0xFF;
	FPipeID.PipeOutId = 0xFF;

	if (WinUsb_QueryInterfaceSettings(FWinUSBHandle, 0, &InterfaceDescriptor))
	{
		for (index = 0; index < InterfaceDescriptor.bNumEndpoints; index++)
		{
			if (WinUsb_QueryPipe(FWinUSBHandle, 0, index, &Pipe)) 
			{
				if (Pipe.PipeType == UsbdPipeTypeBulk)
				{
					if (USB_ENDPOINT_DIRECTION_IN(Pipe.PipeId))
					{
						FPipeID.PipeInId = Pipe.PipeId;
						FReadPacketSize = Pipe.MaximumPacketSize;
					}
					
					if (USB_ENDPOINT_DIRECTION_OUT(Pipe.PipeId))
					{
						FPipeID.PipeOutId = Pipe.PipeId;
						FWritePacketSize = Pipe.MaximumPacketSize;
					}
				}	
			}
		}
	}

	if (FPipeID.PipeInId == 0xFF || FPipeID.PipeOutId == 0xFF) 
	{
		// no bulk pipes available

		WinUSB_CloseDeviceHandles();
		return Error_SetErr(USB_ERR_WR_CREATE_FILE);
	}
	
	if ((FInputThread = WinUSB_CreateInputThread()) == NULL)
	{
		WinUSB_CloseDeviceHandles();
		return false;
	}

	FWriteTimeout = -1;

	// write a null-sized block to tell the device to use WinUSB instead of HID
	WinUsb_SetPipePolicy(FWinUSBHandle, FPipeID.PipeOutId, PIPE_TRANSFER_TIMEOUT, sizeof(timeout), &timeout);
	WinUsb_WritePipe(FWinUSBHandle, FPipeID.PipeOutId, &dummy_buff, 0, &written, 0);

	FThreadRunning = true;
	ResumeThread(FInputThread);

	return true;
}
//---------------------------------------------------------------------------
void WinUSB_SetTimeouts(dword rx_tout, dword tx_tout)
{
	FReadTimeout = rx_tout;
	FWriteTimeout = tx_tout;
}
//---------------------------------------------------------------------------
bool WinUSB_Close(void)
{
	if (FDeviceHandle == INVALID_HANDLE_VALUE)
		return Error_SetErr(USB_ERR_NOT_OPEN);

	if (FStopEvent)
		SetEvent(FStopEvent);

	if (FInputThread)
	{
		// close the input thread's handle
		CloseHandle(FInputThread);
		DeleteCriticalSection(&FQueueLock);
		FInputThread = NULL;
	}

	CloseHandle(FDeviceHandle);
	FDeviceHandle = INVALID_HANDLE_VALUE;

	while (FThreadRunning)
		Sleep(100);

	if (FStopEvent)
	{
		CloseHandle(FStopEvent);
		FStopEvent = NULL;
	}

	WinUSB_CloseDeviceHandles();
	WinUSB_UnLoadDLL();
	return true;
}
//---------------------------------------------------------------------------
bool WinUSB_IsDevOpen(void)
{
	return FDeviceHandle != INVALID_HANDLE_VALUE;
}
//---------------------------------------------------------------------------
bool WinUSB_Transmit(byte *data, DWORD length) 
{
	return WinUSB_TransmitWTimeout(data, length, FWriteTimeout);
}
//---------------------------------------------------------------------------
bool WinUSB_TransmitWTimeout(byte *data, DWORD length, dword timeout)
{
	bool rv;
	dword written;

	if (!WinUSB_IsDevOpen())
		return Error_SetErr(USB_ERR_NOT_OPEN);

	Error_SetErr(0);

	if (!length)
		return true;

	if (FCurrWriteTimeout != (int) timeout)
	{
		rv = WinUsb_SetPipePolicy(FWinUSBHandle, FPipeID.PipeOutId, PIPE_TRANSFER_TIMEOUT, sizeof(timeout), &timeout);
		FWriteTimeout = timeout;

		if (!rv) 
		{
			FWriteTimeout = -1;
			
			return Error_SetErr(USB_ERR_WRITE);
		}
		else
		{
			// printf("\nSetting write timeout successfully to %dms\n", FWriteTimeout);
			FCurrWriteTimeout = timeout;
		}
	}

	rv = WinUsb_WritePipe(FWinUSBHandle, FPipeID.PipeOutId, data, length, &written, 0);

	if (!rv || written != length)
		return Error_SetErr(USB_ERR_WRITE);

	return true;
}
//---------------------------------------------------------------------------

